(** @canonical Dune_lang.Ast *)
module Ast = Dune_lang__Ast

(** @canonical Dune_lang.Atom *)
module Atom = Dune_lang__Atom

(** @canonical Dune_lang.Combinators *)
module Combinators = Dune_lang__Combinators

(** @canonical Dune_lang.Conv *)
module Conv = Dune_lang__Conv

(** @canonical Dune_lang.Cst *)
module Cst = Dune_lang__Cst

(** @canonical Dune_lang.Decoder *)
module Decoder = Dune_lang__Decoder

(** @canonical Dune_lang.Encoder *)
module Encoder = Dune_lang__Encoder

(** @canonical Dune_lang.Escape *)
module Escape = Dune_lang__Escape

(** @canonical Dune_lang.Lexer *)
module Lexer = Dune_lang__Lexer

(** @canonical Dune_lang.Parser *)
module Parser = Dune_lang__Parser

(** @canonical Dune_lang.Syntax *)
module Syntax = Dune_lang__Syntax

(** @canonical Dune_lang.T *)
module T = Dune_lang__T

(** @canonical Dune_lang.Template *)
module Template = Dune_lang__Template

(** @canonical Dune_lang.Versioned_file *)
module Versioned_file = Dune_lang__Versioned_file

(** @canonical Dune_lang.Versioned_file_first_line *)
module Versioned_file_first_line = Dune_lang__Versioned_file_first_line
