(** @canonical Cache.Cache_intf *)
module Cache_intf = Cache__Cache_intf

(** @canonical Cache.Client *)
module Client = Cache__Client

(** @canonical Cache.Key *)
module Key = Cache__Key

(** @canonical Cache.Local *)
module Local = Cache__Local

(** @canonical Cache.Messages *)
module Messages = Cache__Messages

(** @canonical Cache.Messages_intf *)
module Messages_intf = Cache__Messages_intf
